/* * 
 * This file is part of bufferpool
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 * 
 * bufferpool is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * bufferpool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with bufferpool; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA 
 *
 * */

 /**
  * @file bufferpool.h
  * @brief bufferpool simple shared queue
  * 
  * Bufferpool support ONE AND ONLY ONE producer and many consumers.
  * Each consumer have it's own BPConsumer structure that is NOT SHARED
  * with others readers.
  * The only struct that must be mutually exclusive is BPBuffer.
  * */

#ifndef BUFFERPOOL_H
#define BUFFERPOOL_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <stdint.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdlib.h>
#include <stddef.h>
#include <limits.h>
#include <stdarg.h>

#define BPBUFF_MEM_PAGE 9

#define BPBUFFER_DEFAULT_DIM BPBUFF_MEM_PAGE

#define BPSLOT_DATASIZE 2000

#define BPBUFF_SHM_CTRLNAME "Buffer"
#define BPBUFF_SHM_SLOTSNAME "Slots"
#define BPBUFF_SHM_PAGE BPBUFF_MEM_PAGE    // 9

typedef ptrdiff_t BPSlotPtr;

typedef struct {
    uint16_t refs;
    int16_t seq_delta;  /*!< delta to latest sequence number */
    uint64_t slot_seq;  /*!< monotone identifier of slot (NOT RTP seq) */
    double timestamp;   /*!< presentation time of packet */
    double last_timestamp; /*!< presentation time of last packet in buffer */
    uint32_t rtp_time;  //!< if != 0 it contains the calculated rtp timestamp
    uint32_t pkt_num; /*!< number of packet in buffer */
    uint8_t data[BPSLOT_DATASIZE];
    uint32_t data_size;
    uint8_t marker;
    ptrdiff_t next;
} BPSlot;

typedef struct {
    uint16_t refs;        /*!< n.Consumers that share the buffer */
    uint32_t nslots;
    BPSlotPtr write_pos;  /*!< last write position */
    sem_t syn;
} BPControl;

typedef enum { buff_local = 0, buff_shm } BPBufferType;

/*!
 * Buffer struct.
 * the pointers have different meaning if we use normally allocated memory
 * or shared memory.
 * Using a shared memory buffer pointer will be offsets relative to the
 * beginning of SHM.
 * This way the two processes can add the address returned by mmap to obtain
 * the absolute address.
 * */
typedef struct {
    BPBufferType type;       //!< whether buffer is on shared memory or not;
    BPControl *control;
    BPSlot *slots;           /*!< Buffer head */
    uint32_t known_slots;    /*!< last known number of slots.
                                  This member is only useful for SHM buffer. */
    char filename[PATH_MAX]; //!< SHM object file
    uint64_t last_slot_seq;  //!< sequence nuber of last slot inserted in the buffer
    double last_slot_timestamp; //!< timestamp of last slot inserted in the buffer
} BPBuffer;

typedef struct {
    BPSlotPtr read_pos;         /*!< read position */
    BPSlotPtr last_read_pos;    /*!< last read position.
                                    used for managing the slot addition */
    uint64_t last_seq;
    BPBuffer *buffer;
    int32_t frames;
} BPConsumer;

/*! API definitions*/
BPBuffer *bp_new(uint32_t size);
BPConsumer *bp_ref(BPBuffer *buffer);
void bp_unref(BPConsumer *cons);
int bp_read(BPConsumer *cons, uint32_t *timestamp, uint8_t *marker, uint8_t *data,
            uint32_t *data_size);
BPSlot *bp_getreader(BPConsumer *cons);
int bp_gotreader(BPConsumer *cons);
int bp_write(BPBuffer *buffer, int16_t seq_delta, double timestamp, uint32_t rtp_time,
             uint8_t marker, uint8_t *data, uint32_t data_size);
BPSlot *bp_getslot(BPBuffer *buffer);
BPSlot *bp_addpage(BPBuffer *buffer, BPSlot *prev);
//BPSlot *bp_slotadd(BPBuffer *, BPSlot *);
void bp_free(BPBuffer *buffer);

int bp_isempty(BPConsumer *cons);
double bp_nextts(BPConsumer *cons);

// Shared Memory Bufferpool
BPBuffer *bp_shm_create(char *shm_name, uint32_t size);
BPBuffer *bp_shm_map(char *shm_name);
BPSlot *bp_shm_addpage(BPBuffer *buffer);
int bp_shm_remap(BPBuffer *buffer);
// int bp_shm_refresh(BPBuffer *);
#define bp_shm_refresh(oms_buffer) \
    (((oms_buffer->type == buff_shm) && \
      (oms_buffer->known_slots != oms_buffer->control->nslots)) ? \
        bp_shm_remap(oms_buffer) : 0)
int bp_shm_unmap(BPBuffer *buffer);
int bp_shm_destroy(BPBuffer *buffer);

// shared memory names
char *bp_ipc_name(const char *, const char *);

typedef void (*bp_log_t)(int level, const char *fmt, va_list list);

void bp_log_init(bp_log_t fn);

void bp_log(int level, const char *fmt, ...)
#ifdef __GNUC__
  __attribute__((__format__(__printf__, 2, 3)))
#endif
  ;


// BPSlotPtr manipulation
#define BPtoSlot(b, p) \
    ((p<0) ? NULL : (&b->slots[p])) //! used when a pointer could be NULL,
                                    //  otherwise we'll use buffe->slots[index]
#define BPtoSlotPtr(b, p) (p ? p - b->slots : -1)

// dump for debug 
#if ENABLE_DUMPBUFF
void bp_dump(BPConsumer *, BPBuffer *);
#else
#define bp_dump(x, y)
#endif


#endif
